<?php
/**
 * @copyright © 2008, 2009 Intrahealth International, Inc.
 * This File is part of I2CE
 *
 * I2CE is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
*  I2CE_CustomRelationship
* @package I2CE
* @subpackage Core
* @author Carl Leitner <litlfred@ibiblio.org>
* @version 2.1
* @access public
*/


class I2CE_FormRelationship extends I2CE_Fuzzy {



    protected $primaryForm;
    protected $primaryFormName;
    protected $parentFormNames;
    protected $formConfigs;
    protected $relationship;
    protected $relConfig;
    protected $form_reference;


    /**
     * @param string $relationship
     * @param callback $form_reference
     */
    public function __construct($relationship,$form_reference, $relationship_base) {
        $this->relationship = $relationship;
        $this->relConfig = I2CE::getConfig()->traverse("$relationship_base/{$this->relationship}",false,false);
        if (!$this->relConfig instanceof I2CE_MagicDataNode) {
            throw new Exception("Relationship $relationship does not exist under $relationship_base");
        }
        $this->form_reference = $form_reference;
        $this->primaryFormName = $this->relConfig->getName();
        if (!$this->relConfig->setIfIsSet($this->primaryForm,'form')) {
            throw new Exception("No primary form set for $relationship");
        }
        $this->formConfigs = array('primary_form'=>$this->relConfig);
        $this->parentFormNames = array('primary_form' => 'primary_form', $this->relConfig->getName()=>'primary_form');
        $this->getForms($this->relConfig,$this->formConfigs, $this->parentFormNames, 'primary_form');
    }

    public function getForm($formname) {
        $form = false;
        if (!is_string($formname)) {
            return false;
        }
        if (array_key_exists($formname,$this->formConfigs)) {
            $this->formConfigs[$formname]->setIfIsSet($form,'form');
        }
        return $form;
    }
    
    /**
     * Gets the forms required by the relationship
     * @returns array of string
     */
    public function getRequiredForms() {
        $forms = array();
        foreach ($this->formConfigs as $formname=>$formConfig) {
            $form = null;
            if ($formConfig->setIfIsSet($form,'form')) {                
                $forms[] = $form;
            }
        }
        return array_unique($forms);
    }

    protected function getForms($config, &$formConfigs, &$parentForms, $parent_form_name) {
        if (!$config->is_parent('joins')) {
            return;
        }
        foreach ($config->joins as $childForm=>$childConfig) {
            $formConfigs[$childForm] = $childConfig;
            $parentForms[$childForm] = $parent_form_name;
            $this->getForms($childConfig,$formConfigs, $parentForms, $childConfig->getName());
        }
    }


    public function getFormConfig($form = null) {
        if ($form === null) {
            return $this->formConfigs;
        } else {
            if (array_key_exists($formname,$this->formConfigs)) {
                return $this->formConfigs[$form];
            } else {
                return false;
            }
        }
    }

    public function getParentFormNames($formname = null) {
        if ($formname === null) {
            return $this->parentFormNames;
        } else {
            if (array_key_exists($formname,$this->parentFormNames)) {
                return $this->parentFormNames[$formname];
            } else {
                return false;
            }
        }
    }



    public function getPrimaryForm() {
        return $this->primaryForm;
    }

    public function getPrimaryFormName() {
        return $this->primaryFormName;
    }
    public function isPrimaryFormName($formname) {
        return ($this->primaryFormName === $formname);
    }



    public function getFormNames() {
        return array_keys($this->formConfigs);
    }




    public function getReferencedForm($form) {
        $ret = @call_user_func($this->form_reference,$form);
        if (!is_string($form)) {
            I2CE::raiseError("Could not get reference for form $form");
            return false;
        }
        return $ret;
    }

    
    
    public function getFunctionDetails($function) {
        if (!$this->relConfig->is_parent("reporting_functions/$function")) {
            return false;
        }
        $config = $this->relConfig->traverse("reporting_functions/$function");       
        $qry = '';
        if (!$config->setIfIsSet($qry,'qry')) {
            I2CE::raiseError("Function {$function} is not defined");
            return false;
        }
        $qry= trim($qry);
        if (!$qry) { //no query set so skip it
            I2CE::raiseError("Function {$function} is empty");
            return false;
        }
        if (!$config->setIfIsSet($formfield,'formfield') 
            || ! ($fieldObj = I2CE_FormField::createField($config->formfield,$function)) instanceof I2CE_FormField) {
            I2CE::raiseError("Function {$function} cannot be associated to a form field");
            return false;
        }
        preg_match_all('/`([a-zA-Z0-9\_\-]+\+[a-zA-Z0-9\_\-]+)`/',$qry,$required_fields);
        return array('qry'=>$qry,'type'=>$fieldObj->getDBType(),'required_fields'=>array_unique($required_fields[1]), 'field'=>$fieldObj);
    }


    protected static $operands = array('OR','XOR','AND','NOT');

    public function getLimitingFields ($formName) {
        $limiting_fields = array();
        if (!array_key_exists($formName,$this->formConfigs)) {
            return $limiting_fields;
        }
        if (!$this->formConfigs[$formName]->is_parent('where')) {
            return $limiting_fields;
        }
        $this->_getLimitingFields($this->formConfigs[$formName]->where, $limiting_fields);
        return $limiting_fields;
    }


    protected function _getLimitingFields($whereConfig, &$limiting_fields) {
        if (!$whereConfig instanceof I2CE_MagicDataNode) {
            I2CE::raiseError("Invalid where clause");
            return;
        }
        if (!isset($whereConfig->operator)) {
            return;
        }
        if ($whereConfig->operator == 'FIELD_LIMIT') {
            //see if we have a valid field, if so add it to our list
            $whereConfig->setIfIsSet($field,'field');
            if ($field && !in_array($field,$limiting_fields)) {
                $limiting_fields[$field] =true;
            }
        } else if (in_array($whereConfig->operator,self::$operands)) {
            //go down
            if (!$whereConfig->is_parent('operand')) {
                return;
            }
            foreach ($whereConfig->operand as $operand=>$opConfig) {
                $this->_getLimitingFields($opConfig,$limiting_fields);
            }
            
        } //else do nothing
    }


    public function generateWhereClause($formName) {
        if (!array_key_exists($formName,$this->formConfigs) || !$this->formConfigs[$formName]->is_scalar('where/operator')) {
            return '';
        }        
        $form = $this->getForm($formName);
        $formObj = I2CE_FormFactory::instance()->createForm($form.'|0');
        if (!$formObj instanceof I2CE_Form) {
            I2CE::raiseError("Could not instantiate $formName=>$form");
            return false;
        }
        //$callback = create_function('$form,$field',"return \"`$formName+\$field`\";");
        if ( $formName == "primary_form" ) {
            $callback = create_function('$form,$field',"return \"`$form`.`\$field`\";");
        } else {
            $callback = create_function('$form,$field',"return \"`$formName`.`\$field`\";");
        }
        $where =  $formObj->generateWhereClause($this->formConfigs[$formName]->where->getAsArray(),$callback);
        $formObj->cleanup();
        return $where;
    }


    /***
     * 
     * Methods used to make joins
     *
     */

    public function getReferenedForm($form) {
        @$refForm = call_user_function($this->form_reference,$form);
        if (!$refForm) {
            I2CE::raiseError("Could not create reference for $form");
            return false;
        }
        return $refForm;
    }



    protected function joinOn_ids($childForm, $joinData) {
        if ( ($refChildForm = $this->getReferencedForm($this->getForm($childForm))) === false ) {
            return false;
        }
        return " JOIN  $refChildForm AS `$childForm`  ON `$childForm`.id = `parent_form`.id ";
    }

    protected function joinOn_parent($childForm, $joinData) {
        if ( ($refChildForm = $this->getReferencedForm($this->getForm($childForm))) === false ) {
            return false;
        }
        return " JOIN $refChildForm AS `$childForm`  ON `$childForm`.id = `parent_form`.parent ";
    }

    protected function joinOn_child($childForm, $joinData) {
        if ( ($refChildForm = $this->getReferencedForm($this->getForm($childForm))) === false ) {
            return false;
        }
        return " JOIN $refChildForm AS `$childForm`  ON `$childForm`.parent = `parent_form`.id "  ;
    }
    

    protected function joinOn_child_field($childForm, $joinData ) {
        if  ((!is_array($joinData)) || 
             (!array_key_exists('field', $joinData)) || 
             (!is_string($joinData['field'])) ||
             (strlen($joinData['field']) == 0) ){
            I2CE::raiseError("Join child field not specified for parent_form/$childForm");
            return false;
        }
        if ( ($refChildForm = $this->getReferencedForm($this->getForm($childForm))) === false ) {
            return false;
        }
        return  " JOIN $refChildForm AS `$childForm` ON `$childForm`.`{$joinData['field']}` = `parent_form`.id ";
    }
    
    protected function joinOn_parent_field ($childForm, $joinData) {
        if  ((!is_array($joinData)) || 
             (!array_key_exists('field', $joinData)) || 
             (!is_string($joinData['field'])) ||
             (strlen($joinData['field']) == 0) ){
            I2CE::raiseError("Join parent field not specified for parent_form/$childForm");
            return false;
        }
        if ( ($refChildForm = $this->getReferencedForm($this->getForm($childForm))) === false ) {
            return false;
        }
        return " JOIN  $refChildForm AS `$childForm`  ON `$childForm`.id = `parent_form`.`{$joinData['field']}` ";
    }
    
    
    protected function joinOn_fields ($childForm, $joinData) {        
        if  (!is_array($joinData)) {
            I2CE::raiseError("Join data  specified for parent_form/$childForm");
            return false;
        }
        foreach (array('parent','child') as $field) { 
            if ((!array_key_exists($field, $joinData)) || 
                (!is_string($joinData[$field])) ||
                (strlen($joinData[$field]) == 0) ){
                I2CE::raiseError("Join $field field not specified for parent_form/$childForm");
                return false;
            }
        }
        if ( ($refChildForm = $this->getReferencedForm($this->getForm($childForm))) === false ) {
            return false;
        }
        return " JOIN $refChildForm AS `$childForm` ON `$childForm`.`{$joinData['child']}` = `parent_form`.`{$joinData['parent']}` " ;
    }
    

    public function getJoin($formName) {
        if (!array_key_exists($formName, $this->formConfigs)) {
            return false;
        }
        $formConfig = $this->formConfigs[$formName];
        $joinStlye = '';
        $formConfig->setIfisSet($joinStyle,"join_style");
        if (!$joinStyle) {
            I2CE::raiseError("No join style specfied for $formName");
            return false;
        }
        $joinData = array();
        $formConfig->setIfIsSet($joinData,"join_data",true);
        $joinMethod = "joinOn_$joinStyle";
        if (!$this->_hasMethod($joinMethod)) {
            I2CE::raiseError("No method registererd to handle join style {$joinStyle} for $formName");
            return false;
        }
        return $this->$joinMethod($formName, $joinData);
    }
 



}
# Local Variables:
# mode: php
# c-default-style: "bsd"
# indent-tabs-mode: nil
# c-basic-offset: 4
# End:
