<?php
/**
 * @copyright © 2008, 2009 Intrahealth International, Inc.
 * This File is part of I2CE
 *
 * I2CE is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
*  I2CE_SwissConfig_FormRelationship_Join
* @package I2CE
* @subpackage Core
* @author Carl Leitner <litlfred@ibiblio.org>
* @version 2.1
* @access public
*/


class I2CE_Swiss_FormRelationship_Join extends  I2CE_Swiss_FormRelationship {

    public function processValues($vals) {
        $do_style = false;
        $do_field = false;
        if (array_key_exists('drop_empty', $vals) && $vals['drop_empty']) {
            $this->storage->drop_empty = 1;
        } else {
            $this->storage->drop_empty = 0;
        }
        if (array_key_exists('join_style',$vals)) {
            $do_style = true;
            if (!$this->hasParent()) {
                I2CE::raiseError("No parent");
                return false;
            }
            if (!in_array($vals['join_style'], array('parent','child','parent_field','child_field'))) {
                I2CE::raiseError("Invalid join_style:{$vals['join_style']}");
                return false;
            }
            if (in_array($vals['join_style'], array('parent_field','child_field'))){
                $do_field = true;
                if (!array_key_exists('join_field',$vals) || !is_scalar($vals['join_field']) || !$vals['join_field']) {
                    I2CE::raiseError("Invalid joinfield");
                    return false;
                }
                if ($vals['join_style'] == 'parent_field') {
                    $parent = $this->getParent()->getAncestorByClass('I2CE_Swiss_FormRelationship');
                    if (!$parent instanceof I2CE_Swiss_FormRelationship) {
                        I2CE::raiseError("Invalid parent");
                        return false;
                    }
                    $formName = $parent->getForm();
                } else {
                    if (!array_key_exists('form',$vals)) {
                        I2CE::raiseError("No form specified");
                        return false;
                    }
                    $formName = $vals['form'];
                }
                $factory = I2CE_FormFactory::instance();
                $formObj= $factory->createForm($formName);
                if (!$formObj instanceof I2CE_Form) {
                    I2CE::raiseError("Invalid form $formName");
                    return false;
                }                    
                $fields = $factory->getFieldNames($formName,array('in_db'=>true));
                if (!array_key_exists($vals['join_field'],$fields)) {
                    I2CE::raiseError("Invalid field" . $vals['join_field'] . "\n" . print_r($fields,true));
                    return false;
                }
            }
        } 
        if (!parent::processValues($vals)) {
            return false;
        }
        if ($do_field){ 
            $this->storage->join_data->field = $vals['join_field'];
        }
        if ($do_style) {
            $this->storage->join_style = $vals['join_style'];
        }
        return true;
    }


    public function getDescription() {
        $joinDesc = parent::getDescription();
        $joinStyle = '';
        $this->storage->setIfIsSet($joinStyle,'join_style');
        $child = null;
        if (!$this->hasParent()) {
            return $joinDesc;
        }
        $parent = $this->getParent()->getAncestorByClass('I2CE_Swiss_FormRelationship');
        if (!$parent instanceof I2CE_Swiss_FormRelationship) {
            return $joinDesc;
        }
        $parentForm = $parent->getStorage()->getName();
        switch ($joinStyle) {
        case 'parent':
            $joinDesc .= 'This form is a parent of ' . $parentForm . '.';        
            break;
        case 'child':            
            $joinDesc .= 'This form is a child of ' . $parentForm . '.';                
            break;
        case 'child_field':            
            $field = '<NOT SPECIFIED>';
            $this->storage->setIfIsSet($field,"join_data/field");
            $joinDesc .= 'This form a links to ' . $parentForm . ' via one the field ' . $field. '.';               
            break;
        case 'parent_field':
            $field = '<NOT SPECIFIED>';
            $this->storage->setIfIsSet($field,"join_data/field");
            $joinDesc .= 'This form is  linked from ' . $parentForm . ' via the field '  . $field. '.';                
            break;
        default:
            break;
        }
        return $joinDesc;
    }


    /**
     * Displays the meta data about this form
     * @param mixed$configPath
     * @param mixed $configPath
     * @param DOMNode $appendNode The page template node from which we wish to add the join menu to
     * @returns boolean true  on success
     */
    protected function displayMetaData( $action,$node ) {
        return parent::displayMetaData('view',$node);        
    }


    protected function showDropEmpty($action,$contentNode) {
        $file = 'formRelationship_join_drop_empty_' . $action . '.html';
        $dropNode = $this->template->appendFileByNode($file,'div',$contentNode);        
        if (!$dropNode instanceof DOMNode) {
            I2CE::raiseError("Could not load template for drop data");
            return false;
        }
        $drop = 0;
        $this->storage->setIfIsSet($drop,"drop_empty");
        if ($action == 'edit') {
            if ($drop) {
                $id = 'do_drop_empty';
            } else {
                $id = 'no_drop_empty';
            }
            $radio = $this->template->getElementById($id,$dropNode);
            if ($radio instanceof DOMElement) {
                $radio->setAttribute('checked','checked');
            }
            $this->renameInputs(array('drop_empty'),$dropNode);
        } else { //view
            $this->template->setDisplayDataImmidiate('do_drop_empty',$drop,$dropNode);
        }
    }

    protected function manageFunctions($action, $node) {
        $this->template->setDisplayDataImmediate('reporting_functions_link', '', $node);
    }



}
# Local Variables:
# mode: php
# c-default-style: "bsd"
# indent-tabs-mode: nil
# c-basic-offset: 4
# End:
