<?php
/**
* © Copyright 2009 IntraHealth International, Inc.
* 
* This File is part of I2CE 
* 
* I2CE is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License 
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
* @package i2ce
* @subpackage forms
* @author Carl Leitner <litlfred@ibiblio.org>
* @version v4.0.0
* @since v4.0.0
* @filesource 
*/ 
/** 
* Class I2CE_FormStorage_CSV
* 
* @access public
*/


class I2CE_FormStorage_CSV extends   I2CE_FormStorage_Mechanism{
    /**
     * Construct this module class
     * @param string $name The name of this storage mechanism
     * @param array $options
     */
    public function __construct( $name, $options=array() ) {
        parent::__construct( $name, $options );

        if (array_key_exists('closeCSV',$options)) {
            $this->closeCSV = $opions['closeCSV'];
        }
    }

    /**
     * whether or not to close the file pointer for a CSV between access.
     * @var protected boolean $closeCSV 
     */
    protected $closeCSV = false;




    /**
     * @var protected array $resources.  An array indexed by form names of file pointers.
     */
    protected $resources = array(); 

    /**
     *  Tries to get the filepointer  for the CSV file
     * @param string $form 
     * @returns resource
     */
    protected function getCSV($form) {
        if (array_key_exists($form,$this->resources) && is_resource($this->resources[$form])) {
            return $this->resources[$form];
        }
        $options = $this->getStorageOptions($form);
        if (!$options instanceof I2CE_MagicDataNode) {
            I2CE::raiseError("Invalid CSV storage options for $form");
            return false;
        }
        $file = null;
        if (!$options->setIfIsSet($file,"file")) {
            I2CE::raiseError("CSV file is not set");
            return false;
        }
        if (!I2CE_FileSearch::isAbsolut($file)) {
            $csv_file = I2CE::getFileSearch()->search( 'CSV', $file );
        } else {
            $csv_file = $file;
        }
        if (!$csv_file || !is_readable($csv_file)) {
            I2CE::raiseError("Could not find CSV $file ($csv_file)");
            return false;
        }
        $fp = fopen($csv_file,"r");
        if (!is_resource($fp)) {
            I2CE::raiseError("Could not open $csv_file for reading");
            return false;
        }
        $this->resources[$form] = $fp;
        return $this->resources[$form];
    }


    /**
     *  Tries to get the filepointer  for the CSV file
     * @param string $form 
     * @returns resource
     */
    protected function closeCSV($form) {
        if (array_key_exists($form,$this->resources) && is_resource($this->resources[$form])) {
            fclose($this->resources[$form]);
            $this->resources[$form] = false;
        }
    }
    

    /**
     * Array of file offesset for an id for a given form.
     */
    protected $locations  = array();


    /**
     * Array of file last read offets for a given form.
     */
    protected $lastread_locations  = array();




    /**
     * Get the offset in the CSV file for the specified form object
     * @param string $formName
     * @params string $id
     * @returns int on success, false on failure.
     */
    protected function getFileLocation($formName,$id) {
        if (!array_key_exists($formName,$this->locations) || !is_array($this->locations[$formName])) {
            $this->locations[$formName] = array();
        }
        if (!array_key_exists($id,$this->locations[$formName])) {
            $this->getFileLocations($this->getCSV($formName),$formName,$id);
        }
        return $this->locations[$formName][$id];
    }

    /**
     * Worker moethod to  the offset in the CSV file for the specified form id and store it into the {$locations} cache.
     * Reads through the file until the specified is found.
     * @param resource $p
     * @param string $form
     * @param string $id.  If false we get all of the file locations
     * @returns int on success, false on failure.
     */
    protected function getFileLocations($fp,$form,$id) {
        if (!array_key_exists($form,$this->locations) || !is_array($this->locations[$form])) {
            $this->locations[$form] = array();
        }
        if (!is_resource($fp)) {
            if ($id !== false) {
                $this->locations[$form][$id] = false;
            }
            return false;
        }
        $last_read = 0;
        $storageOptions = $this->getStorageOptions($form);
        if (!$storageOptions instanceof I2CE_MagicDataNode) {
            I2CE::raiseError("Bad storage options");
            if ($id !== false) {
                $this->locations[$form][$id] = false;
            }
            return false;
        }
        $has_header = true;
        $storageOptions->setIfIsSet($has_header,'has_header');
        $last_read = 0;
        if (array_key_exists($form,$this->lastread_locations)) {
            $last_read = $this->lastread_locations[$form];
        }
        if ($has_header && $last_read == 0) {
            //skip the header row
            if ( fseek($fp,0) != 0) {
                I2CE::raiseError("Could not seek to header");
                if ($id !== false) {
                    $this->locations[$form][$id] = false;
                }
                return false;
            }
            if (fgets($fp) === false) {
                I2CE::raiseError("Could not read line");
                if ($id !== false) {
                    $this->locations[$form][$id] = false;
                }
                return false;
            }
            $this->lastread_locations[$form] = ftell($fp);
        } else {
            if ( fseek($fp,$last_read) != 0) {
                I2CE::raiseError("Could not seek to header");
                if ($id !== false) {
                    $this->locations[$form][$id] = false;
                }
                return false;
            }            
        }
        $start_read = ftell($fp);
        $found = false;
        while ($data = $this->getCSVData($fp,$form)) {            
            $this->locations[$form][$data['id']] = $start_read;
            if ($id !== false && $data['id'] == $id) {
                $found = true;
                break;
            }
            $start_read =ftell($fp);
        }
        $this->lastread_locations[$form] = ftell($fp);
        if (!$found && $id !== false) {
            $this->locations[$form][$id] = false;
        }
        if ($id !== false) {
            return $this->locations[$form][$id] = false;
        } else {
            return true;
        }
    }


    /**
     * Gets the form data for a specified form as an array of readable and enabled datavalues from the CSV indexed by the field names, 'id' and 'parent'
     * @param string $form
     * @param string $id
     * @returns array or false on failure
     */
    protected function getFormData($form,$id) {
        if (!is_resource($fp = $this->getCSV($form))) {
            I2CE::raiseError( "No saved data found for form " , $form);
            return false;
        }            
        $location = $this->getFileLocation($form,$id);
        if ($location === false) {
            //this id is not saved in the CSV. don't worry about it.
            return false;
        }
        if (fseek($fp,$location) != 0) {
            I2CE::raiseError("Seek error for $form|$id");
            return false;
        }        
        $data = $this->getCSVData($fp,$form);
        if ($data['id'] != $id) {
            return false;
        }
        return $data;
    }


    /**
     * Reads a line of data from ta CSV file pointer and returns it as an array of readable and enabled datavalues from the CSV indexed by the field names, 'id' and 'parent'
     * @param resource $fp
     * @param string $form
     * @returns mixed array on success, false on failure
     */
    protected function getCSVData($fp,$form) {
        $storageOptions = $this->getStorageOptions($form);
        if (!$storageOptions instanceof I2CE_MagicDataNode) {
            I2CE::raiseError("Bad storage options");
            return false;
        }
        $delimiter =',';
        $storageOptions->setIfIsSet($delimiter,'delimiter');
        $enclosure = '"';
        $storageOptions->setIfIsSet($enclosure,'enclosure');
        $form_prepended = true;
        $storageOptions->setIfIsSet($form_prepended,'id/form_prepended');
        $parent = true;
        $storageOptions->setIfIsSet($parent,'parent/enabled');
        $data = fgetcsv($fp, 0, $delimiter, $enclosure);
        if (!is_array($data)) {
            return false;
        }
        if (!$this->ensureIndices($fp,$form)) {
            return false;
        }
        $ret = array();
        foreach ($this->field_indices[$form] as $field=>$index) {
            if (array_key_exists($index,$data)) {
                $ret[$field] = $data[$index];
            } else {
                $ret[$field] = null;
            }
        }
        if (!array_key_exists('id',$ret)) {
            //the id was not found.
            return false;
        }
        $pos = strpos($ret['id'],'|');
        if ($form_prepended) {
            if ($pos === false) {
                //bad id.
                return false; 
            } else {
                $ret['id'] = substr($ret['id'],$pos+1);
            }
        }
        if (!$parent || !array_key_exists('parent',$ret) || $ret['parent'] === null) {
            $ret['parent'] = '0';
        }
        return $ret;
    }

    /**
     * The column indices for the fields
     * @var protected array $field_indices.  Arrays indexed by form names of pairs $field=>$index
     */
    protected $field_indices = array();

    /**
     * Ensures that the column indices for the fields are set.
     * @param resource $fp
     * @param string $form
     * @returns true on success
     */
    protected function ensureIndices($fp,$form) {
        if (array_key_exists($form,$this->field_indices)) {
            return is_array($this->field_indices[$form]);
        }
        $storageOptions = $this->getStorageOptions($form);
        if (!$storageOptions instanceof I2CE_MagicDataNode) {
            $this->field_indices[$form] = false;
            return false;
        }
        $factory = I2CE_FormFactory::instance();
        $formObj = $factory->createForm($form);
        if (!$formObj instanceof I2CE_Form) {
            I2CE::raiseError("Could not instantiate $form");
            $this->field_indices[$form] = false;
            return false;
        }
        $use_header = true;
        $has_header = true;
        $delimiter =',';
        $enclosure = '"';
        $storageOptions->setIfIsSet($use_header,'use_header');
        $storageOptions->setIfIsSet($has_header,'has_header');
        $storageOptions->setIfIsSet($delimiter,'delimiter');
        $storageOptions->setIfIsSet($enclosure,'enclosure');
        $fields = array('parent'=>'parent','id'=>'id');
        foreach ($formObj->getFieldNames() as $field) {
            $fields[$field] = "fields/$field";
        }
        $formObj->cleanup();
        $indices = array();
        if ($use_header && $has_header) {
            if (($fp_loc = ftell($fp)) === false) {
                I2CE::raiseError("could not get current location");
                $this->field_indices[$form] = false;
                return false;
            }
            if (fseek($fp,0) === -1) {
                I2CE::raiseError("could not seek to beginining of CSV");                
                $this->field_indices[$form] = false;
                return false;
            }
            $headers = fgetcsv($fp, 0, $delimiter, $enclosure);
            if (!is_array($headers)) {
                I2CE::raiseError("Could not get headers");
                $this->field_indices[$form] = false;
                return false;
            }
            foreach ($headers as &$header) {
                $header = strtolower(trim($header));
            }
            unset($header);
            foreach ($fields as $field=>$path) {
                $enabled = true;
                $storageOptions->setIfIsSet($enabled,$path . '/' . $enabled);
                if (!$enabled) {
                    continue;
                }
                $header = $field;
                $storageOptions->setIfIsSet($header,$path . '/' . $header);
                $index = array_search(strtolower(trim($header)), $headers);
                if ($index !== false) {
                    $indices[$field] = $index;
                }
            }
            if (fseek($fp,$fp_loc) === -1) {
                I2CE::raiseError("could not seek");
                $this->field_indices[$form] = false;
                return false;
            }            
        } else { //don't use the header... indices are integers
            foreach ($fields as $field=>$path) {
                $enabled = true;
                $storageOptions->setIfIsSet($enabled,$path . '/' . $enabled);
                if (!$enabled) {
                    continue;
                }
                $index = false;
                if ($storageOptions->setIfSet($index, $path . '/' . $index)) {
                    $indices[$field] = $index;
                }
            }
        }
        if (!array_key_exists('id',$indices)) {
            I2CE::raiseError("Id column not set");
            $this->field_indices[$form] = false;
            return false;
        }            
        $this->field_indices[$form] = $indices;
        return true;
    }


    /**********************************
     * 
     * Implement the abstract read methods
     *
     **********************************


    
    /**
     * Checks to see if this is writalbe
     * @returns boolean
     */
    public function isWritable() {
        return false;
    }



    /**
     * Return an array of all the record ids for a given form.
     * @param string $form
     * @return array
     */
    public function getRecords($form) {
        $this->getFileLocations($this->getCSV($form),$form,false); //get all id's
        if ($this->closeCSV) {
            $this->closeCSV($form->getForm());
        }
        return array_keys($this->locations[$form]);
    }
    

    /**
     * Populate the member variables of the object from the database.
     * @param I2CE_Form $form
     */
    public function populate( $form) {
        $formName = $form->form();
        $data = $this->getFormData($form->form(),$form->getId());
        if ($this->closeCSV) {
            $this->closeCSV($form->getForm());
        }
        if (!is_array($data)) {
            //not is CSV file so assume it is OK
            return true;
        }
        foreach ($data as $field=>$dbval) {
            if ($field == 'id') {
                //do nothing
            } else if ($field == 'parent') {
                if ($dbval != '0') {
                    $form->setParent($dbval);
                }
            } else {
                $fieldObj = $form->getField($field);
                if (!$fieldObj instanceof I2CE_FormField) {
                    continue;
                }
                $fieldObj->setFromDB($dbval);
            }
        }
        return true;
    }


    
    

}
# Local Variables:
# mode: php
# c-default-style: "bsd"
# indent-tabs-mode: nil
# c-basic-offset: 4
# End:
